package org.eclipse.osgi.util;

import java.util.Properties;
import java.io.InputStream;
import java.io.IOException;
import java.util.HashMap;
import java.lang.reflect.Field;
import java.util.Map;
import java.util.ArrayList;
import java.util.Locale;
import java.security.AccessController;
import java.security.PrivilegedAction;

public abstract class NLS
{
    private static final Object[] EMPTY_ARGS;
    private static final String EXTENSION = ".properties";
    private static String[] nlSuffixes;
    static final int SEVERITY_ERROR = 4;
    static final int SEVERITY_WARNING = 2;
    static final Object ASSIGNED;
    
    static {
        EMPTY_ARGS = new Object[0];
        ASSIGNED = new Object();
    }
    
    protected NLS() {
    }
    
    public static String bind(final String message, final Object binding) {
        return internalBind(message, null, String.valueOf(binding), null);
    }
    
    public static String bind(final String message, final Object binding1, final Object binding2) {
        return internalBind(message, null, String.valueOf(binding1), String.valueOf(binding2));
    }
    
    public static String bind(final String message, final Object[] bindings) {
        return internalBind(message, bindings, null, null);
    }
    
    public static void initializeMessages(final String bundleName, final Class clazz) {
        if (System.getSecurityManager() == null) {
            load(bundleName, clazz);
            return;
        }
        AccessController.doPrivileged((PrivilegedAction<Object>)new PrivilegedAction() {
            public Object run() {
                NLS.load(bundleName, clazz);
                return null;
            }
        });
    }
    
    private static String internalBind(final String message, Object[] args, final String argZero, final String argOne) {
        if (message == null) {
            return "No message available.";
        }
        if (args == null || args.length == 0) {
            args = NLS.EMPTY_ARGS;
        }
        final int length = message.length();
        int bufLen = length + args.length * 5;
        if (argZero != null) {
            bufLen += argZero.length() - 3;
        }
        if (argOne != null) {
            bufLen += argOne.length() - 3;
        }
        final StringBuffer buffer = new StringBuffer((bufLen < 0) ? 0 : bufLen);
        for (int i = 0; i < length; ++i) {
            final char c = message.charAt(i);
            switch (c) {
                case '{': {
                    final int index = message.indexOf(125, i);
                    if (index == -1) {
                        buffer.append(c);
                        break;
                    }
                    if (++i >= length) {
                        buffer.append(c);
                        break;
                    }
                    int number = -1;
                    try {
                        number = Integer.parseInt(message.substring(i, index));
                    }
                    catch (final NumberFormatException ex) {
                        throw new IllegalArgumentException();
                    }
                    if (number == 0 && argZero != null) {
                        buffer.append(argZero);
                    }
                    else if (number == 1 && argOne != null) {
                        buffer.append(argOne);
                    }
                    else {
                        if (number >= args.length || number < 0) {
                            buffer.append("<missing argument>");
                            i = index;
                            break;
                        }
                        buffer.append(args[number]);
                    }
                    i = index;
                    break;
                }
                case '\'': {
                    final int nextIndex = i + 1;
                    if (nextIndex >= length) {
                        buffer.append(c);
                        break;
                    }
                    final char next = message.charAt(nextIndex);
                    if (next == '\'') {
                        ++i;
                        buffer.append(c);
                        break;
                    }
                    final int index = message.indexOf(39, nextIndex);
                    if (index == -1) {
                        buffer.append(c);
                        break;
                    }
                    buffer.append(message.substring(nextIndex, index));
                    i = index;
                    break;
                }
                default: {
                    buffer.append(c);
                    break;
                }
            }
        }
        return buffer.toString();
    }
    
    private static String[] buildVariants(String root) {
        if (NLS.nlSuffixes == null) {
            String nl = Locale.getDefault().toString();
            final ArrayList result = new ArrayList(4);
            while (true) {
                result.add(String.valueOf('_') + nl + ".properties");
                final int lastSeparator = nl.lastIndexOf(95);
                if (lastSeparator == -1) {
                    break;
                }
                nl = nl.substring(0, lastSeparator);
            }
            result.add(".properties");
            NLS.nlSuffixes = result.toArray(new String[result.size()]);
        }
        root = root.replace('.', '/');
        final String[] variants = new String[NLS.nlSuffixes.length];
        for (int i = 0; i < variants.length; ++i) {
            variants[i] = String.valueOf(root) + NLS.nlSuffixes[i];
        }
        return variants;
    }
    
    private static void computeMissingMessages(final String bundleName, final Class clazz, final Map fieldMap, final Field[] fieldArray, final boolean isAccessible) {
        for (final Field field : fieldArray) {
            if ((field.getModifiers() & 0x19) == 0x9) {
                if (fieldMap.get(field.getName()) != NLS.ASSIGNED) {
                    try {
                        final String value = "NLS missing message: " + field.getName() + " in: " + bundleName;
                        log(2, value, null);
                        if (!isAccessible) {
                            field.setAccessible(true);
                        }
                        field.set(null, value);
                    }
                    catch (final Exception e) {
                        log(4, "Error setting the missing message value for: " + field.getName(), e);
                    }
                }
            }
        }
    }
    
    static void load(final String bundleName, final Class clazz) {
        System.currentTimeMillis();
        final Field[] fieldArray = clazz.getDeclaredFields();
        final ClassLoader loader = clazz.getClassLoader();
        final boolean isAccessible = (clazz.getModifiers() & 0x1) != 0x0;
        final int len = fieldArray.length;
        final Map fields = new HashMap(len * 2);
        for (int i = 0; i < len; ++i) {
            fields.put(fieldArray[i].getName(), fieldArray[i]);
        }
        final String[] variants = buildVariants(bundleName);
        for (int j = 0; j < variants.length; ++j) {
            final InputStream input = (loader == null) ? ClassLoader.getSystemResourceAsStream(variants[j]) : loader.getResourceAsStream(variants[j]);
            if (input != null) {
                try {
                    final MessagesProperties properties = new MessagesProperties(fields, bundleName, isAccessible);
                    properties.load(input);
                }
                catch (final IOException e) {
                    log(4, "Error loading " + variants[j], e);
                    if (input != null) {
                        try {
                            input.close();
                        }
                        catch (final IOException ex) {}
                    }
                    continue;
                }
                finally {
                    if (input != null) {
                        try {
                            input.close();
                        }
                        catch (final IOException ex2) {}
                    }
                }
                if (input != null) {
                    try {
                        input.close();
                    }
                    catch (final IOException ex3) {}
                }
            }
        }
        computeMissingMessages(bundleName, clazz, fields, fieldArray, isAccessible);
    }
    
    static void log(final int severity, final String message, final Exception e) {
        String statusMsg = null;
        switch (severity) {
            case 4: {
                statusMsg = "Error: ";
                break;
            }
            default: {
                statusMsg = "Warning: ";
                break;
            }
        }
        if (message != null) {
            statusMsg = String.valueOf(statusMsg) + message;
        }
        if (e != null) {
            statusMsg = String.valueOf(statusMsg) + ": " + e.getMessage();
        }
        System.err.println(statusMsg);
        if (e != null) {
            e.printStackTrace();
        }
    }
    
    private static class MessagesProperties extends Properties
    {
        private static final int MOD_EXPECTED = 9;
        private static final int MOD_MASK = 25;
        private static final long serialVersionUID = 1L;
        private final String bundleName;
        private final Map fields;
        private final boolean isAccessible;
        
        public MessagesProperties(final Map fieldMap, final String bundleName, final boolean isAccessible) {
            this.fields = fieldMap;
            this.bundleName = bundleName;
            this.isAccessible = isAccessible;
        }
        
        @Override
        public synchronized Object put(final Object key, final Object value) {
            final Object fieldObject = this.fields.put(key, NLS.ASSIGNED);
            if (fieldObject == NLS.ASSIGNED) {
                return null;
            }
            if (fieldObject == null) {
                final String msg = "NLS unused message: " + key + " in: " + this.bundleName;
                NLS.log(2, msg, null);
                return null;
            }
            final Field field = (Field)fieldObject;
            if ((field.getModifiers() & 0x19) != 0x9) {
                return null;
            }
            try {
                if (!this.isAccessible) {
                    field.setAccessible(true);
                }
                field.set(null, value);
            }
            catch (final Exception e) {
                NLS.log(4, "Exception setting field value.", e);
            }
            return null;
        }
    }
}
